(*
 * SNU 4541.664A Program Analysis 2017 Spring
 * K- Interpreter Solution
 * DongKwon Lee (dklee@ropas.snu.ac.kr)
 *)


(*
 * K- Interpreter
 *)

module type KMINUS =
sig
  exception Error of string
  type id = string
  type exp =
      | NUM of int
      | VAR of id
	  | DEREF of id
	  | AT of id
      | ADD of exp * exp
	  | MINUS of exp
	  | TRUE
	  | FALSE
	  | NOT of exp
	  | ANDALSO of exp * exp
	  | ORELSE of exp * exp
	  | LESS of exp * exp
  
  type cmd =
      | SKIP
      | SEQ of cmd * cmd        (* sequence *)
      | IF of exp * cmd * cmd   (* if-then-else *)
	  | WHILE of exp * cmd      (* while *)
      | ASSIGN of id * exp      (* assgin to variable *)
	  | PTRASSIGN of id * exp

  type program = cmd
  type memory
  type value

  val botMemory : memory
  val eval : memory * exp -> value  (* exp eval *)
  val assume : memory * exp -> memory (* memory filtering *)
  val analyzer : memory * program -> memory
  val used_varlist : program -> id list
  val pp_memory : memory -> id list -> unit
end

module K : KMINUS =
struct
  exception Error of string
  type id = string
  type exp =
      | NUM of int
      | VAR of id
	  | DEREF of id
	  | AT of id
      | ADD of exp * exp
	  | MINUS of exp
	  | TRUE
	  | FALSE
	  | NOT of exp
	  | ANDALSO of exp * exp
	  | ORELSE of exp * exp
	  | LESS of exp * exp
  
  type cmd =
      | SKIP
      | SEQ of cmd * cmd        (* sequence *)
      | IF of exp * cmd * cmd   (* if-then-else *)
	  | WHILE of exp * cmd      (* while *)
      | ASSIGN of id * exp      (* assgin to variable *)
      | PTRASSIGN of id * exp


  type program = cmd
  
  (* abstract domain type *)
  type z_hat = BOT_Z | CONST of int | TOP_Z
  type bool_hat = TOP_BOOL | T | F | BOT_BOOL
  type loc_hat = LSET of string list

  (* abstract memory, value type *)
  type value = z_hat * bool_hat * loc_hat
  type memory = id -> value
  
  (* list functions *)
  let rec list_trim l =
    match l with
	| [] -> []
	| hd::tl -> if (List.mem hd tl) then list_trim tl else hd::(list_trim tl)

  let rec list_max zlist =
    match zlist with
	| [] -> 0
	| hd::[] -> hd
	| hd::tl -> if(hd < list_max(tl)) then list_max(tl) else hd

  let rec list_min zlist =
    match zlist with
	| [] -> 0
	| hd::[] -> hd
	| hd::tl -> if(hd < list_min(tl)) then hd else list_min(tl)


  (* bottom values *)
  let botValue = (BOT_Z, BOT_BOOL, LSET [])
  let botMemory = fun x -> botValue
  
  (* memory operation *)
  let store mem id v = fun x -> if (x = id) then v else mem(x)
  
  let rec compare_mem m1 m2 varlist =
    match varlist with
	| [] -> true
	| hd::tl -> (m1(hd) = m2(hd)) && (compare_mem m1 m2 tl)
 
  (* join operaters *)
  let join_z z1 z2 =
    match (z1, z2) with
	| (TOP_Z, _) | (_, TOP_Z) -> TOP_Z
	| (BOT_Z, _) -> z2
	| (_, BOT_Z) -> z1
	| (CONST z1, CONST z2) -> if (z1 = z2) then CONST z1 else TOP_Z

  let join_bool b1 b2 = 
    match (b1, b2) with
	| (TOP_BOOL, _) | (_, TOP_BOOL) -> TOP_BOOL
	| (BOT_BOOL, _) -> b2  
	| (_, BOT_BOOL) -> b1
    | (T, F) | (F, T) -> TOP_BOOL
	| (T, T) -> T
	| (F, F) -> F

  let join_loc l1 l2 =
    match (l1, l2) with
	| (LSET llist1, LSET llist2) -> LSET (list_trim (llist1@llist2))

  let join_val v1 v2 =
    let (z1, b1, l1) = v1 in
	let (z2, b2, l2) = v2 in
    ((join_z z1 z2), (join_bool b1 b2), (join_loc l1 l2))

  let join_mem m1 m2 = fun x ->(join_val (m1(x)) (m2(x)))
 

  (* var list gathering : "do not" change here *)
  let rec used_vars_exp exp =
      (match exp with
      | NUM i -> []
      | VAR id | DEREF id | AT id -> id::[]
	  | ADD (e1, e2) -> (used_vars_exp e1) @ (used_vars_exp e2)
	  | MINUS e -> used_vars_exp e
	  | TRUE -> []
	  | FALSE -> []
	  | NOT e -> used_vars_exp e
	  | ANDALSO (e1, e2) -> (used_vars_exp e1) @ (used_vars_exp e2)
	  | ORELSE (e1, e2) -> (used_vars_exp e1) @ (used_vars_exp e2)
	  | LESS (e1, e2) -> (used_vars_exp e1) @ (used_vars_exp e2)
	  )
  
  let rec used_vars cmd =
	match cmd with
	| SKIP -> []
	| SEQ (cmd1, cmd2) -> (used_vars cmd1) @ (used_vars cmd2)
	| IF (e, cmd1, cmd2) -> (used_vars_exp e) @ (used_vars cmd1) @ (used_vars cmd2)
	| WHILE (e, cmd) -> (used_vars_exp e) @ (used_vars cmd)
	| ASSIGN (id, e) -> id::(used_vars_exp e)
	| PTRASSIGN (id, e) -> id::(used_vars_exp e)

  let used_varlist cmd = list_trim (used_vars cmd)


  (* exp evaluation *)
  let rec eval : (memory * exp) -> value  = fun (mem, e) ->
    match e with
	| NUM i -> (CONST i, BOT_BOOL, LSET [])
	| VAR x -> mem(x)
    | DEREF x -> 
      let (z, b, l) = mem(x) in
	  let LSET llist = l in
	  let rec llist_join llist =
	    (match llist with
		| [] ->botValue
		| hd::tl -> join_val (mem(hd)) (llist_join tl)
		) in
	  llist_join llist
    | AT x -> (BOT_Z, BOT_BOOL, LSET (x::[]))
	| ADD (e1, e2) ->
	  let (z1, _, _) = eval (mem, e1) in
	  let (z2, _, _) = eval (mem, e2) in
	  let newz =
	    (match (z1, z2) with
	    | (TOP_Z, _) | (_, TOP_Z) -> TOP_Z
		| (BOT_Z, _) | (_, BOT_Z) -> BOT_Z
	    | (CONST z1, CONST z2) -> CONST (z1 + z2)
        ) in
	  (newz, BOT_BOOL, LSET [])
	| MINUS e -> 
	  let (z, _, _) = eval (mem, e) in 
	  let newz =
	    (match z with
		| TOP_Z -> TOP_Z
		| BOT_Z -> BOT_Z
		| CONST z -> CONST (-1 * z)
		) in
	  (newz, BOT_BOOL, LSET[])
	| TRUE -> (BOT_Z, T, LSET [])
	| FALSE -> (BOT_Z, F, LSET [])
	| NOT e -> 
	  let (_, b, _) = eval (mem, e) in 
	  let newb =
	    (match b with
	    | TOP_BOOL -> TOP_BOOL
	    | T -> F
	    | F -> T
	    | BOT_BOOL -> BOT_BOOL
		) in
      (BOT_Z, newb, LSET [])
	| ANDALSO (e1, e2) ->
	  let (_, b1, _) = eval (mem, e1) in
	  let (_, b2, _) = eval (mem, e2) in
	  let newb =
	    (match (b1, b2) with
		| (BOT_BOOL, _) | (_, BOT_BOOL) -> BOT_BOOL
		| (F, _) | (_, F) -> F
		| (T, _) -> b2
		| (_, T) -> b1
		| (_, _) -> TOP_BOOL
		) in
	  (BOT_Z, newb, LSET [])
	| ORELSE (e1, e2) ->
	  let (_, b1, _) = eval (mem, e1) in
	  let (_, b2, _) = eval (mem, e2) in
	  let newb =
	    (match (b1, b2) with
		| (BOT_BOOL, _) | (_, BOT_BOOL) -> BOT_BOOL
		| (T, _) | (_, T) -> T
		| (F, _) -> b2
		| (_, F) -> b1
		| (_, _) -> TOP_BOOL 
		) in
	  (BOT_Z, newb, LSET [])
	| LESS (e1, e2) -> 
	  let (z1, _, _) = eval (mem, e1) in
	  let (z2, _, _) = eval (mem, e2) in
	  let newb =
	    (match (z1, z2) with
		| (TOP_Z, _) | (_, TOP_Z) -> TOP_BOOL
		| (BOT_Z, _) | (_, BOT_Z) -> BOT_BOOL
		| (CONST z1, CONST z2) -> if (z1 < z2) then T else F
		) in
	  (BOT_Z, newb, LSET [])
	 
  (* memory filtering by boolean expression *)
  let rec assumeNot : (memory * exp) -> memory = fun (mem, e) ->
    let (_, tb, _) = eval(mem, e) in
	  match tb with
	  | BOT_BOOL -> botMemory
	  | T -> botMemory
	  | F -> mem
	  | TOP_BOOL -> (
    match e with
	| NUM i -> botMemory
	| VAR id -> 
	  let (_, b, _) = mem(id) in 
	  (match b with
	  | BOT_BOOL -> botMemory
	  | T -> botMemory
	  | F -> mem
	  | TOP_BOOL -> store mem id (BOT_Z, F, LSET [])
	  )
	| DEREF id -> 
      let (z, b, LSET llist) = mem(id) in
	  let rec llist_selectF llist =
	    (match llist with
		| [] -> []
		| hd::tl ->
		  let (_, hdb, _) = mem(hd) in
		  match hdb with
		  | BOT_BOOL -> (llist_selectF tl)
		  | T -> (llist_selectF tl)
		  | F -> hd::(llist_selectF tl)
		  | TOP_BOOL -> hd::(llist_selectF tl)
		) in
	  let newval = (BOT_Z, BOT_BOOL, LSET (llist_selectF llist)) in
	  store mem id newval
	| AT id -> botMemory
	| ADD (e1, e2) -> botMemory
	| MINUS e -> botMemory
	| TRUE -> botMemory
	| FALSE -> mem
	| NOT e -> assume (mem, e)
	| ANDALSO (e1, e2) -> join_mem (assumeNot(mem, e1)) (assumeNot(mem, e2))
	| ORELSE (e1, e2) -> assumeNot (assumeNot(mem, e1), e2)
    | LESS (e1, e2) -> 
	  let (z1, _, _) = eval (mem, e1) in
	  let (z2, _, _) = eval (mem, e2) in
      (match (z1, z2) with
	  | (BOT_Z, _) | (_, BOT_Z) -> botMemory
	  | (TOP_Z, _) | (_, TOP_Z) -> mem
	  | (CONST z1, CONST z2) -> if (z2 <= z1) then mem else botMemory
	  )
	)
  and assume : (memory * exp) -> memory = fun (mem, e) ->
    let (_, tb, _) = eval (mem, e) in
	  match tb with
	  | BOT_BOOL -> botMemory
	  | T -> mem
	  | F -> botMemory
	  | TOP_BOOL -> (
    match e with
	| NUM i -> botMemory
	| VAR id -> 
	  let (_, b, _) = mem(id) in 
	  (match b with
	  | BOT_BOOL -> botMemory
	  | T -> mem
	  | F -> botMemory
	  | TOP_BOOL -> store mem id (BOT_Z, T, LSET []))
	| DEREF id -> 
      let (z, b, LSET llist) = mem(id) in
	  let rec llist_selectT llist =
	    (match llist with
		| [] -> []
		| hd::tl ->
		  let (_, hdb, _) = mem(hd) in
		  match hdb with
		  | BOT_BOOL -> (llist_selectT tl)
		  | T -> hd::(llist_selectT tl)
		  | F -> (llist_selectT tl)
		  | TOP_BOOL -> hd::(llist_selectT tl)
		) in
	  let newval = (BOT_Z, BOT_BOOL, LSET (llist_selectT llist)) in
	  store mem id newval
	| AT id -> botMemory
	| ADD (e1, e2) -> botMemory
	| MINUS e -> botMemory
	| TRUE -> mem
	| FALSE -> botMemory
	| NOT e -> assumeNot (mem, e)
	| ANDALSO (e1, e2) -> assume (assume(mem, e1), e2)
	| ORELSE (e1, e2) -> join_mem (assume(mem, e1)) (assume(mem, e2))
    | LESS (e1, e2) -> (* workplace *)
	  let (z1, _, _) = eval (mem, e1) in
	  let (z2, _, _) = eval (mem, e2) in
      (match (z1, z2) with
	  | (BOT_Z, _) | (_, BOT_Z) -> botMemory
	  | (TOP_Z, _) | (_, TOP_Z) -> mem
	  | (CONST z1, CONST z2) -> if (z1 < z2) then mem else botMemory
	  )
	)
 


  (* memory pretty print : "do not" change here *)
  let rec pp_zhat : z_hat -> unit = fun z ->
    match z with
    | TOP_Z -> print_string("TOP")
    | CONST i -> print_int(i)
	| BOT_Z -> print_string("bottom")
 
 let rec pp_lochat : loc_hat -> unit = fun loc ->
    match loc with 
    | LSET [] -> ()
    | LSET (hd::tl) -> let _ = print_string(hd ^ ", ") in (pp_lochat (LSET tl)) 


  let pp_bool : bool_hat -> unit = fun b -> 
    match b with 
    | BOT_BOOL -> print_string("bottom") 
    | T -> print_string("T")
    | F -> print_string("F")
    | TOP_BOOL -> print_string("T, F")

  let rec pp_memory : memory -> id list -> unit = fun mem -> (fun varlist ->
    match varlist with 
    | [] -> ()
    | hd::tl ->
      let (zhat, b, lochat) = mem(hd) in
      let _ = print_string(hd ^ " -> Z : ") in
      let _ = pp_zhat(zhat) in
      let _ = print_string(" bool : ") in
      let _ = pp_bool(b) in
      let _ = print_string(" loc : [ ") in
      let _ = pp_lochat(lochat) in
      let _ = print_string("]") in
      let _ = print_newline() in
      pp_memory mem tl
      )



  (* interval analysis for K- *)
  let rec analyzer : (memory * program) -> memory = fun (mem, pgm) ->
    let varlist = used_varlist pgm in
	match pgm with
	| SKIP -> mem
    | SEQ(cmd1, cmd2) -> let mem1 = analyzer (mem, cmd1) in analyzer (mem1, cmd2) 
    | IF(e, cmd1, cmd2) -> 
	  let mem1 = analyzer (assume(mem, e) , cmd1) in
	  let mem2 = analyzer (assume(mem, NOT e) , cmd2) in 
	  join_mem mem1 mem2
	| WHILE(e, cmd) ->
	  let fix_iter = fun m -> join_mem mem (analyzer (assume (m, e), cmd)) in
	  let rec fixpoint_helper m =    
	    if (compare_mem (join_mem m (fix_iter m)) m varlist) then m else fixpoint_helper (join_mem m (fix_iter m)) in
	  let fixpoint = fixpoint_helper botMemory in
	  assume (fixpoint, NOT e)
    | ASSIGN(id, e) -> 
	  let v = eval (mem, e) in
	  (fun x -> if (x = id) then v else mem(x))
 	| PTRASSIGN (id, e) ->
	  let v = eval (mem, e) in
	  let (_, _, LSET llist) = mem(id) in
	  let rec ptrassign_helper mem llist =
	    (match llist with
		| [] -> mem
		| hd::tl -> 
		  let originval = mem(hd) in
		  let newval = join_val originval v in
		  let newmem = store mem hd newval in
		  ptrassign_helper newmem tl
		) in
	  ptrassign_helper mem llist
end
